<?php
include_once dirname(__FILE__).'/../lib/Stato/i18n/i18n.php';
include_once dirname(__FILE__).'/of_http.php';

/**
 * This class handles requests to obm-sync services
 * To implement a service callable on obm-sync, the name of the service must be
 * added to the implementedServices property and a method must be implemented.
 *
 * For a service named "foo", a method getFooUrl() must be implemented and "foo"
 * must be added to the implementedServices array.
 *
 * You can implement a getFooHttpHeaders() returning an array of headers if some
 * specific headers need to be sent for the foo webservice.
 *
 * You can implement a getFooPostParameters() returning an array of parameters
 * if some specific parameters need to be sent for the foo webservice.
 *
 * You can implement a getErrorMessageForFoo() to determine if there is an error
 * depending on the xml response. If not implemented, an error message will only
 * occur if there is an error tag in xml response to webservice or if there is
 * no response.
 */
class obmSyncRequester{
  private       $errors                 = array();
  private       $sid                    = false;
  private       $serverIp               = null;
  private       $authKind              = null;
  private       $login;
  private       $password;
  private       $domainName;
  private       $rootPath               = null;
  private       $defaultRequestHeaders  = array(
                                            "Content-Type",
                                            "application/x-www-form-urlencoded; charset=utf-8"
                                          );
  private       $implementedServices    =  array("login", "importICalendar");
  private       $lastXmlValidResponse   = null;
  private       $httpRequester;
  public static $rootPathScheme         = "http://%s:8080";
  public static $origin                 = "obm-ui";
  const         LOGIN_PATH              = "/obm-sync/services/login/doLogin";

  /**
   * Initializes mandatory properties for services calls
   *
   * @param int $domainId     Domain id in database
   * @param string $authKind  Authentification method
   * @param HTTPRequester     HttpRequester which will be responsible of doing
   *                          HTTP requests to obm-sync
   * @param string $origin    Origin of the request
   */
  public function __construct($serverIp,
                              $authKind,
                              HTTPRequester $httpRequester,
                              $origin = null) {
    $this->httpRequester = $httpRequester;
    $this->authKind = $authKind;
    if(!$serverIp){
      throw new Exception("Invalid obm-sync server ip", 500);
    }
    else{
      $this->serverIp = $serverIp;
      $this->rootPath = sprintf(self::$rootPathScheme, $this->serverIp);
    }
  }

  /**
   * Request headers setter
   * @param array $requestHeaders Must be an $key=>$values array where keys are
   *              header names and values = header values
   */
  public function setDefaultRequestHeaders(array $requestHeaders){
    if(is_array($requestHeaders)){
      $this->defaultRequestHeaders = $requestHeaders;
    }
  }

  private function getLoginHttpHeaders(){
    if($this->authKind == "LemonLDAP"){
      if(!$this->login || !$this->domainName){
        throw new Exception("You must set login and domain name before trying to log in with LemonLDAP", 500);
      }
      return array("obm_uid" => $this->login, "obm_domain" => $this->domainName);
    }
    return array();
  }

  /**
   * Calculates obm-sync login url
   * Must not return an empty string in any case. An exception must be thrown if
   * an element is missing
   * 
   * @return <string> obm-sync login webservice url
   */
  private function getLoginUrl(){
    if(!$this->serverIp){
      throw new Exception("Login failed : no obm-sync server defined in obmSyncRequester", 500);
    }
    if(!$this->rootPath){
      throw new Exception("Login failed : no obm-sync path defined in obmSyncRequester", 500);
    }

    $loginAuthParameters = "?origin=".self::$origin;
    if($this->authKind == "standalone"){
      if(!$this->login){
        throw new Exception("For a standalone authentication, you must set the login and password in obmSyncRequester", 500);
      }
    }
    if($authKind != "LemonLDAP"){
      if($this->login)
        $loginAuthParameters .= "&login=".$this->login;
      if($this->password)
        $loginAuthParameters .= "&password=".$this->password;
    }
    return $this->rootPath."/obm-sync/services/login/doLogin".$loginAuthParameters;
  }

  private function getImportICalendarUrl(){
    $insertPath = "/obm-sync/services/calendar/importICalendar";
    $insertUrl = $this->rootPath.$insertPath."?sid=".$this->sid."&calendar=".$this->login;

    return $insertUrl;
  }

  private function getImportICalendarPostParameters($icsFile){
    return array("ics"=>$icsFile);
  }

  /**
   * Executes obmSync login and gets a session id. Must be executed before
   * calling a service from obm-sync.
   * 
   * @return boolean True if everything went well, false otherwise
   */
  public function obmSyncLogin(){
    try {
      $loginUrl = $this->getLoginUrl();
      $response = $this->httpRequester
                  ->executeHTTPRequest(
                    $loginUrl,
                    $this->getLoginHttpHeaders()
                  );
      if($response === false){
        $this->errors[] = __("Couldn't login to obm-sync");
        throw new Exception("No response from login request", 500);
      }

      // XML parsing in order to find some eventual errors
      $xmlLogin = new DomDocument();
      $xmlLogin->loadXML($response);
      $errors = $xmlLogin->getElementsByTagName("error");
      if($errors->length > 0){
        $this->errors[] = __( "Couldn't login to obm-sync (%error%)",
                              array("error"=>$errors->item(0)->nodeValue)
                            );
        throw new Exception("An error occured at login request", 500);
      }
    } catch (Exception $exc) {
      return false;
    }

    // Everything went well : user is logged in and has a session id
    $this->sid = $xmlLogin->getElementsByTagName('sid')
                  ->item(0)
                  ->nodeValue;
    return true;
  }

  /**
   * Tells if a service is available or not
   * To be available, a service must have a method to define its url and must
   * be defined in implementedServices
   *
   * @param string $serviceName
   * @return boolean True if service exists, false otherwise
   */
  private function serviceExists($serviceName, $getServiceUrlMethod){
    $serviceMethodExists = method_exists($this, $getServiceUrlMethod);
    return in_array($serviceName, $this->implementedServices) && $serviceMethodExists;
  }

  private function getServiceUrlMethod($serviceName){
    return "get".ucfirst($serviceName)."Url";
  }

  private function getHttpHeadersForService($serviceName, $arguments = array()){
    $methodName = "get".ucfirst($serviceName)."HttpHeaders";
    if(method_exists($this, $methodName)){
      return call_user_func_array(array($this, $methodName), $arguments);
    }
    return array();
  }

  private function getPostParametersForService($serviceName, $arguments = array()){
    $methodName = "get".ucfirst($serviceName)."PostParameters";
    if(method_exists($this, $methodName)){
      return call_user_func_array(array($this, $methodName), $arguments);
    }
    return array();
  }

  /**
   * Retrieves eventual error message from an obm-sync webservice response
   *
   * @param mixed Response to an obm-sync webservice after a http request
   * @param string $serviceName Name of the called service
   * @return string Error message if exists, empty string otherwise
   */
  private function getErrorMessageFromResponse($response, $serviceName){
    $methodName = "getErrorMessageFor".  ucfirst($serviceName);
    if(!$response){
      if(method_exists($this, $methodName))
        return call_user_func_array(array($this, $methodName), array());
      else
        return $this->getDefaultErrorMessage($serviceName);
    }
    $xmlResponse = new DomDocument();
    $xmlResponse->loadXML($response);
    $errors = $xmlResponse->getElementsByTagName("error");
    if($errors->length > 0){
      if(method_exists($this, $methodName)){
        return call_user_func_array(array($this, $method_name), array($xmlResponse, $errors));
      }
      else{
        return $this->getDefaultErrorMessage($serviceName, $xmlResponse, $errors);
      }
    }
    $this->lastXmlValidResponse = $xmlResponse;
    return null;
  }

  /**
   * Return an error message if there is no response or if an <error> tag exists
   * 
   * @param string $serviceName
   * @param DOMDocument $response
   * @param DOMNodeList $errors
   * @return string error message
   */
  private function getDefaultErrorMessage($serviceName, DOMDocument $response = null, DOMNodeList $errors = null){
    if(!$response){
      return __(
              "Nothing returned from service '%service_name%' call",
              array("service_name"=>$serviceName)
             );
    }
    if($errors instanceof DOMNodeList && $errors->length > 0){
      $error = $errors->item(0)->nodeValue ?
             $errors->item(0)->nodeValue :
             __("unknown error");
      return __(
              "An error occured when calling service '%service_name%' : %error description%",
              array(
                "service_name" => $serviceName,
                "error description" => $error)
              );
    }
    return false;
  }

  private function getErrorMessageForImportICalendar(DOMDocument $response = null, DOMNodeList $errors = null){
    if(!$response){
      return __("An unknown error occured during the ics import");
    }
    $error = $errors->item(0)->nodeValue ?
             $errors->item(0)->nodeValue :
             __("An unknown error occured during the ics import");
    if(preg_match("/Calendar : duplicate with same extId found/", $error, $matches)){
      preg_match("/\[.*\]/", $error, $matches);

      $error = __("An event from the ics already exists")." ".$matches[0];
    }

    return $error;
  }

  /**
   * Calls an obm-sync service
   * 
   * @param string $serviceName
   * @return mixed returns the xml of the response as a DomDocument or false if
   *               an error occured
   */
  public function callObmSyncService($serviceName, $arguments = array()){
    // Login if not already done
    if(!$this->sid){
      if(!$this->obmSyncLogin()){
        return false;
      }
    }

    $getServiceUrlMethod = $this->getServiceUrlMethod($serviceName);
    if(!$this->serviceExists($serviceName, $getServiceUrlMethod)){
      throw new Exception("The obm-sync service '".$serviceName."' isn't defined");
    }
    $httpHeaders =  array_merge(
                      $this->defaultRequestHeaders,
                      $this->getHttpHeadersForService($serviceName, $arguments)
                    );
    $postParameters = $this->getPostParametersForService($serviceName, $arguments);
    $serviceUrl = call_user_func_array(array($this, $getServiceUrlMethod), $arguments);
    $response = $this->httpRequester
                ->executeHTTPRequest(
                  $serviceUrl,
                  $httpHeaders,
                  $postParameters
                );

    if($errorMessage = $this->getErrorMessageFromResponse($response, $serviceName)){
      $this->errors[] = $errorMessage;
      return false;
    }

    return $this->lastXmlValidResponse;
  }

  public function getErrors(){
    return $this->errors;
  }

  public function countErrors(){
    return count($this->errors);
  }

  public function setLogin($login){
    $this->login = $login;
  }

  public function setPassword($password){
    $this->password = $password;
  }

  public function setDomainName($domainName){
    $this->domainName = $domainName;
  }

  public function isLoggedIn(){
    return $this->sid != null;
  }

  public function getLastValidXmlResponse(){
    return $this->lastXmlValidResponse;
  }

  public function getFirstError(){
    if($this->countErrors()){
      $errors = $this->errors;
      return $errors[0];
    }
    else{
      return null;
    }
  }
}
?>
